"""Tut.

Usage:
  tut init [<path>]
  tut points
  tut checkpoint <name> [-m <message>]
  tut edit <name>

Options:
  -h --help     Show this screen.
  --version     Show version.

"""

from __future__ import absolute_import
import os
import sys

from docopt import docopt

import tut
from tut.model import (
    Tut,
    TutException,
)


def init(args):

    path = args.get('<path>')
    if path is None:
        path = os.getcwd()
    else:
        path = os.path.join(os.getcwd(), path)

    tut_repo = Tut(path)

    if not os.path.exists(os.path.join(path, '.git')):
        tut_repo.init()
    tut_repo.install_hooks()


def points(args):

    for point in Tut(os.getcwd()).points():
        print point


def checkpoint(args):

    point = Tut(os.getcwd()).checkpoint(
        args['<name>'],
        message=args.get('<message>'),
    )

    print "Recorded checkpoint: %s" % point


def edit(args):

    Tut(os.getcwd()).edit(
        args['<name>'],
    )


CMD_MAP = {
    'init': init,
    'points': points,
    'checkpoint': checkpoint,
    'edit': edit,
}


def main():
    arguments = docopt(__doc__, version='Tut %s' % tut.version())

    for cmd in CMD_MAP:
        if arguments.get(cmd):
            try:
                CMD_MAP[cmd](arguments)
            except TutException, e:
                print "Error: %s" % e
                sys.exit(1)

            break


def post_rewrite():

    tut = Tut(os.getcwd())

    for line in sys.stdin:
        rewrite = line.split()
        tut.move_checkpoints(rewrite[0].strip(), rewrite[1].strip())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    Sphinx test suite utilities
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: Copyright 2007-2013 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.

    This is largely lifted directly from the Sphinx repository,
    because the Sphinx test directory isn't included with the
    installable package.

    The ``with_sphinx`` function has been improved to use a kward
    instead of the first positional argument for the Sphinx
    Application, which allows it to work with TestCase methods.

"""

import sys
import StringIO
import tempfile
import shutil
import re
from codecs import open

try:
    from functools import wraps
except ImportError:
    # functools is new in 2.4
    wraps = lambda f: (lambda w: w)

from sphinx import application
from sphinx.theming import Theme
from sphinx.ext.autodoc import AutoDirective

from path import path

# from nose import tools, SkipTest


__all__ = [
    'test_root', 'test_roots', 'raises', 'raises_msg',
    'skip_if', 'skip_unless', 'skip_unless_importable', 'Struct',
    'ListOutput', 'TestApp', 'with_app', 'gen_with_app',
    'path', 'with_tempdir', 'write_file',
    'sprint', 'remove_unicode_literals',
]


test_root = path(__file__).parent.joinpath('root').abspath()
test_roots = path(__file__).parent.joinpath('roots').abspath()


def _excstr(exc):
    if type(exc) is tuple:
        return str(tuple(map(_excstr, exc)))
    return exc.__name__

def raises(exc, func, *args, **kwds):
    """
    Raise :exc:`AssertionError` if ``func(*args, **kwds)`` does not
    raise *exc*.
    """
    try:
        func(*args, **kwds)
    except exc:
        pass
    else:
        raise AssertionError('%s did not raise %s' %
                             (func.__name__, _excstr(exc)))

def raises_msg(exc, msg, func, *args, **kwds):
    """
    Raise :exc:`AssertionError` if ``func(*args, **kwds)`` does not
    raise *exc*, and check if the message contains *msg*.
    """
    try:
        func(*args, **kwds)
    except exc, err:
        assert msg in str(err), "\"%s\" not in \"%s\"" % (msg, err)
    else:
        raise AssertionError('%s did not raise %s' %
                             (func.__name__, _excstr(exc)))

def skip_if(condition, msg=None):
    """Decorator to skip test if condition is true."""
    def deco(test):
        @tools.make_decorator(test)
        def skipper(*args, **kwds):
            if condition:
                raise SkipTest(msg or 'conditional skip')
            return test(*args, **kwds)
        return skipper
    return deco

def skip_unless(condition, msg=None):
    """Decorator to skip test if condition is false."""
    return skip_if(not condition, msg)

def skip_unless_importable(module, msg=None):
    """Decorator to skip test if module is not importable."""
    try:
        __import__(module)
    except ImportError:
        return skip_if(True, msg)
    else:
        return skip_if(False, msg)


class Struct(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class ListOutput(object):
    """
    File-like object that collects written text in a list.
    """
    def __init__(self, name):
        self.name = name
        self.content = []

    def reset(self):
        del self.content[:]

    def write(self, text):
        self.content.append(text)


class TestApp(application.Sphinx):
    """
    A subclass of :class:`Sphinx` that runs on the test root, with some
    better default values for the initialization parameters.
    """

    def __init__(self, srcdir=None, confdir=None, outdir=None, doctreedir=None,
                 buildername='html', confoverrides=None,
                 status=None, warning=None, freshenv=None,
                 warningiserror=None, tags=None,
                 confname='conf.py', cleanenv=False):

        application.CONFIG_FILENAME = confname

        self.cleanup_trees = [test_root / 'generated']

        if srcdir is None:
            srcdir = test_root
        if srcdir == '(temp)':
            tempdir = path(tempfile.mkdtemp())
            self.cleanup_trees.append(tempdir)
            temproot = tempdir / 'root'
            test_root.copytree(temproot)
            srcdir = temproot
        elif srcdir == '(empty)':
            tempdir = path(tempfile.mkdtemp())
            self.cleanup_trees.append(tempdir)
            temproot = tempdir / 'root'
            temproot.makedirs()
            (temproot / 'conf.py').write_text('')
            srcdir = temproot
        else:
            srcdir = path(srcdir)
        self.builddir = srcdir.joinpath('_build')
        self.cleanup_trees.append(self.builddir)
        if confdir is None:
            confdir = srcdir
        if outdir is None:
            outdir = srcdir.joinpath(self.builddir, buildername)
            if not outdir.isdir():
                outdir.makedirs()
            self.cleanup_trees.insert(0, outdir)
        if doctreedir is None:
            doctreedir = srcdir.joinpath(srcdir, self.builddir, 'doctrees')
            if not doctreedir.isdir():
                doctreedir.makedirs()
            if cleanenv:
                self.cleanup_trees.insert(0, doctreedir)
        if confoverrides is None:
            confoverrides = {}
        if status is None:
            status = StringIO.StringIO()
        if warning is None:
            warning = ListOutput('stderr')
        if freshenv is None:
            freshenv = False
        if warningiserror is None:
            warningiserror = False

        application.Sphinx.__init__(self, srcdir, confdir, outdir, doctreedir,
                                    buildername, confoverrides, status, warning,
                                    freshenv, warningiserror, tags)

    def cleanup(self, doctrees=False):
        Theme.themes.clear()
        AutoDirective._registry.clear()
        for tree in self.cleanup_trees:
            shutil.rmtree(tree, True)


def with_sphinx(*args, **kwargs):
    """
    Make a TestApp with args and kwargs, pass it to the test and clean up
    properly.
    """

    def generator(target):

        if isinstance(target, type):
            # decorating a class, look for test methods to decorate
            for name, prop in target.__dict__.items():
                if name.startswith('test'):
                    setattr(target, name, generator(prop))

            return target

        @wraps(target)
        def deco(*args2, **kwargs2):
            app = TestApp(*args, **kwargs)
            try:
                target(*args2, sphinx_app=app, **kwargs2)
            finally:
                app.cleanup()
        return deco
    return generator


def gen_with_app(*args, **kwargs):
    """
    Decorate a test generator to pass a TestApp as the first argument to the
    test generator when it's executed.
    """
    def generator(func):
        @wraps(func)
        def deco(*args2, **kwargs2):
            app = TestApp(*args, **kwargs)
            for item in func(app, *args2, **kwargs2):
                yield item
            # don't execute cleanup if test failed
            app.cleanup()
        return deco
    return generator


def with_tempdir(func):
    def new_func(*args, **kwds):
        tempdir = path(tempfile.mkdtemp())
        func(tempdir, *args, **kwds)
        tempdir.rmtree()
    new_func.__name__ = func.__name__
    return new_func


def write_file(name, contents, encoding=None):
    if encoding is None:
        mode = 'wb'
        if isinstance(contents, unicode):
            contents = contents.encode('ascii')
    else:
        mode = 'w'
    f = open(str(name), mode, encoding=encoding)
    f.write(contents)
    f.close()


def sprint(*args):
    sys.stderr.write(' '.join(map(str, args)) + '\n')

_unicode_literals_re = re.compile(r'u(".*?")|u(\'.*?\')')
def remove_unicode_literals(s):
    return _unicode_literals_re.sub(lambda x: x.group(1) or x.group(2), s)



import os
import sys

from sh import (
    git,
    which,
)


BRANCH_PREFIX = '_tut_edit_'


def branch_name(tag_or_point):

    return BRANCH_PREFIX + tag_or_point


class TutException(Exception):
    pass


class Tut(object):

    def __init__(self, path):
        self.__path = path

    def _git_dir(self):
        """Return the actual location of the git dir for the repo.

        For submodules, this may be nested in a parent .git directory.

        """

        dotgit = os.path.join(self.__path, '.git')
        assert os.path.exists(dotgit)

        if os.path.isdir(dotgit):
            return dotgit

        # .git is a file, which means we're probably in a submodule.
        with file(dotgit, 'r') as gitfile:
            for line in gitfile:
                if line.startswith('gitdir: '):
                    return os.path.join(
                        self.__path,
                        line.strip().split('gitdir: ', 1)[1]
                    )

        raise TutException("Could not determine the location of the git repo.")

    def _repo_dirty(self):
        """Return True if the repository is dirty."""

        UNCHANGED_STATUSES = (' ', '?', '!')

        for status_line in git.status(porcelain=True):
            if not(status_line[0] in UNCHANGED_STATUSES and
                   status_line[1] in UNCHANGED_STATUSES):
                return True

        return False

    def _hooks_installed(self):
        """Return True if the tut hooks are installed in the Repo."""

        # right now just make sure some post-rewrite hook is installed
        return os.path.exists(
            os.path.join(self._git_dir(), 'hooks', 'post-rewrite')
        )

    def init(self):
        """Create a new repository with an initial commit."""

        cwd = os.getcwd()

        git.init(self.__path)

        os.chdir(self.__path)
        git.commit(
            m='Initializing empty Tut project.',
            allow_empty=True,
        )

        os.chdir(cwd)

    def install_hooks(self):
        """Install git hooks for tut."""

        # install the special post-rewrite hook
        hook_target = which('tut_remap')
        if not hook_target:
            # look for it alongside ourselves
            if os.path.exists(
                    os.path.join(
                        os.path.dirname(sys.argv[0]),
                        'tut_remap',
                    )
            ):
                hook_target = os.path.join(
                    os.path.dirname(sys.argv[0]),
                    'tut_remap',
                )

        if hook_target:
            os.symlink(
                os.path.abspath(hook_target),
                os.path.join(self._git_dir(), 'hooks', 'post-rewrite'),
            )

    def points(self):
        """Return a list of existing checkpoints (tags)."""

        tut_points = [
            tag.strip() for tag in git.tag(
                # '-n1',  # maybe n=1 would work, too
                _iter=True,
            )
        ]

        return tut_points

    def _branches(self):

        return [
            b.split()[-1].strip()
            for b in git.branch('-q')
        ]

    def _editting(self, name):
        """Return True if an edit is in progress for point_name."""

        return branch_name(name) in self._branches()

    def _commit_all_and_tag(self, message, tag_name, force=False):
        """Commit all changes in the working tree and tag with tag_name."""

        # commit all changes
        git.commit(
            all=True,
            m=message,
        )

        # tag the checkpoint
        git.tag(tag_name, force=force)

    def checkpoint(self, name, message=None):
        """Create a new checkpoint, or apply edits to an existing one."""

        if name in self.points():
            if self._editting(name):
                return self.finish_edit(name, message)
            else:
                raise TutException("Point already exists")

        if message is None:
            # generate a default message
            message = 'Tut Checkpoint: %s' % name

        self._commit_all_and_tag(message, name)

    def edit(self, name):
        """Start editing the checkpoint point_name."""

        if not self._hooks_installed():
            raise TutException(
                "Tut hook not installed; do you need to run tut init?"
            )

        # make sure this is a known checkpoint
        if name not in self.points():
            raise TutException("Unknown checkpoint.")

        # make sure the repo is clean
        if self._repo_dirty():
            raise TutException("Dirty tree.")

        # create a new branch to contain editing
        git.checkout(
            name,
            b=branch_name(name),
        )

    def finish_edit(self, name, message=None):
        """Complete an edit in progress and rebase any future points."""

        if message is None:
            # generate a default message
            message = 'Tut Checkpoint: %s' % name

        self._commit_all_and_tag(message, name, force=True)

        # rebase master onto our branch
        git.rebase(
            branch_name(name),
            'master',
        )

        # delete the editing branch
        git.branch('-d', branch_name(name))

    def move_checkpoints(self, old_sha, new_sha):
        """Move any checkpoints that point at old_sha to new_sha."""

        for tag_name in git.tag('--points-at=%s' % old_sha, _iter=True):
            tag_name = tag_name.strip()

            print git.tag(
                tag_name,
                new_sha,
                force=True,
            )



import os
from unittest import TestCase

from mock import patch

import tut.sphinx
from tut.tests import util
from tut.tests.util import with_sphinx


@patch('os.chdir', new=lambda x:None)
@patch('tut.sphinx.git', return_value='original_branch')
class SphinxExtensionLifecycleTests(TestCase):

    # the order of these decorators is *important*: cleanup needs to
    # be replaced before the Sphinx Test Application is instantiated
    @patch('tut.sphinx.cleanup')
    @with_sphinx()
    def test_cleanup_called(self, cleanup_mock, git_mock, sphinx_app=None):

        sphinx_app.build(force_all=True)

        self.assertEqual(cleanup_mock.call_count, 1)

    @with_sphinx()
    def test_cleanup_resets_paths(self, git_mock, sphinx_app=None):

        start_dir = os.getcwd()

        sphinx_app.build()

        self.assertEqual(git_mock.checkout.call_count, 2)
        self.assertEqual(git_mock.checkout.call_args_list[0][0],
                         ('step_one',))
        self.assertEqual(git_mock.checkout.call_args_list[1][0],
                         ('original_branch',))

        self.assertEqual(start_dir, os.getcwd())


@patch('os.chdir', new=lambda x:None)
@patch('tut.sphinx.git')
class TutDirectiveTests(TestCase):

    @with_sphinx(srcdir=util.test_root.parent/'tut_directive')
    def test_set_default_path(self, git_mock, sphinx_app=None):
        """tut directive sets the default repo path."""

        self.assertEqual(tut.sphinx._DEFAULT_PATH, None)
        sphinx_app.builder.build_all()

        # check the resolved paths in the cache
        self.assert_(
            os.path.join(
                os.path.dirname(__file__), 'tut_directive', 'src'
            ) in tut.sphinx._RESET_PATHS
        )
        self.assertEqual(tut.sphinx._DEFAULT_PATH, '/src')

        # cleanup
        sphinx_app.cleanup()
        sphinx_app.builder.cleanup()


@patch('os.chdir', new=lambda x:None)
@patch('tut.sphinx.git')
class CheckpointDirectiveTests(TestCase):

    @with_sphinx()
    def test_checkpoint_triggers_checkout(self, git_mock, sphinx_app=None):
        sphinx_app.build()

        self.assertEqual(git_mock.checkout.call_count, 2)
        self.assertEqual(git_mock.checkout.call_args_list[0][0],
                         ('step_one',))

    @with_sphinx()
    def test_checkpoint_resets_pycode_cache(self, git_mock, sphinx_app=None):

        # dirty the cache
        import sphinx.pycode
        sphinx.pycode.ModuleAnalyzer.cache = {'foo': 'bar'}

        # build the project, which includes a checkout directive
        sphinx_app.build()

        self.assertEqual(sphinx.pycode.ModuleAnalyzer.cache, {})

    @with_sphinx(srcdir=util.test_root.parent/'relative_path')
    def test_relative_paths(self, git_mock, sphinx_app=None):
        """Paths are relative to the document location."""

        sphinx_app.builder.build_all()

        # check the resolved paths in the cache
        self.assert_(
            os.path.join(
                os.path.dirname(__file__), 'relative_path', 'src'
            ) in tut.sphinx._RESET_PATHS
        )

        # cleanup
        sphinx_app.cleanup()
        sphinx_app.builder.cleanup()

    @with_sphinx(srcdir=util.test_root.parent/'abs_path')
    def test_absolute_paths(self, git_mock, sphinx_app=None):
        """Absolute paths are relative to the project root."""

        sphinx_app.builder.build_all()

        # check the resolved paths in the cache
        self.assert_(
            os.path.join(
                os.path.dirname(__file__), 'abs_path', 'src'
            ) in tut.sphinx._RESET_PATHS
        )

        # cleanup
        sphinx_app.cleanup()
        sphinx_app.builder.cleanup()



from __future__ import absolute_import
import os

from sh import git
from docutils.parsers.rst import Directive, directives
import sphinx.pycode


_DEFAULT_PATH = None
_RESET_PATHS = {}


class TutDefaults(Directive):
    option_spec = {
        'path': directives.path,
    }

    def run(self):

        global _DEFAULT_PATH
        _DEFAULT_PATH = self.options['path']

        return []


class TutCheckpoint(Directive):

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        'path': directives.path,
    }

    def run(self):

        global _DEFAULT_PATH
        global _RESET_PATHS

        if 'path' in self.options:
            tut_path = self.options['path']
        elif _DEFAULT_PATH is not None:
            tut_path = _DEFAULT_PATH
        else:
            raise Exception("No tut path specified.")

        # paths are relative to the project root
        rel_path, tut_path = self.state.document.settings.env.relfn2path(
            tut_path)

        curdir = os.getcwd()
        os.chdir(tut_path)

        # if this is the first time visiting this repo
        if tut_path not in _RESET_PATHS:
            # record the current branch
            _RESET_PATHS[tut_path] = \
                git('name-rev', 'HEAD').strip().split()[-1]

        git.checkout(self.arguments[0].strip().lower())
        sphinx.pycode.ModuleAnalyzer.cache = {}

        os.chdir(curdir)

        return []


def initialize(app):

    global _RESET_PATHS
    _RESET_PATHS = {}


def cleanup(app, exception):

    global _RESET_PATHS

    curdir = os.getcwd()
    try:
        for path in _RESET_PATHS:
            os.chdir(path)
            git.checkout(_RESET_PATHS[path])
    finally:
        os.chdir(curdir)


def setup(app):

    app.add_directive('tut', TutDefaults)
    app.add_directive('checkpoint', TutCheckpoint)

    app.connect('builder-inited', initialize)
    app.connect('build-finished', cleanup)



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.1'

install_requires = [
    'docopt',
    'sh',
    'Sphinx',
]


setup(name='tut',
      version=version,
      description="",
      long_description=README + '\n\n' + NEWS,
      classifiers=[
      ],
      keywords='',
      author='Nathan Yergler',
      author_email='nathan@yergler.net',
      url='http://github.com/nyergler/tut',
      license='BSD',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      entry_points={
          'console_scripts': [
              'tut=tut.cmd:main',
              'tut_remap=tut.cmd:post_rewrite',
          ],
      },
)
